use parking_lot;

use serenity;
use serenity::client::CACHE;
use serenity::model::id::{ChannelId, GuildId};
use serenity::voice::{AudioSource, Handler, LockedAudio};

use std::cell::RefCell;
use std::ops::DerefMut;
use std::sync::Arc;
use std::thread;
use std::time;

macro_rules! check_option {
    ($v:expr) => {
        match $v {
            Some(t) => t,
            None => return (),
        }
    };
}

pub struct Song {
    pub source: Box<AudioSource>,
    pub name: String,
}

pub struct MediaData {
    playlist: RefCell<Vec<Song>>,
    current_song: RefCell<Option<LockedAudio>>,
}

fn guild_id(channel_id: ChannelId) -> Option<GuildId> {
    match CACHE.read().guild_channel(channel_id) {
        Some(channel) => Some(channel.read().guild_id),
        None => {
            super::check_msg(channel_id.say("Error finding channel info"));
            None
        }
    }
}

fn handler<'a>(
    guild_id: GuildId,
    manager: &'a mut parking_lot::MutexGuard<
        '_,
        serenity::client::bridge::voice::ClientVoiceManager,
    >,
) -> Option<&'a mut Handler> {
    manager.get_mut(guild_id)
}

impl MediaData {
    fn start_thread(
        media: &Arc<MediaData>,
        channel_id: ChannelId,
        manager_lock: &Arc<
            serenity::prelude::Mutex<serenity::client::bridge::voice::ClientVoiceManager>,
        >,
    ) {
        let media_clone = media.clone();
        let manager_clone = manager_lock.clone();

        thread::spawn(move || loop {
            let mut borrow = media_clone.current_song.borrow_mut();

            if let Some(ref mut song) = borrow.deref_mut() {
                let song_lock = song.clone();
                let audio = song_lock.lock();

                if audio.finished {
                    Self::next_song(
                        song,
                        media_clone.playlist.borrow_mut().deref_mut(),
                        channel_id,
                        &manager_clone,
                    )
                }

                continue;
            }

            let mut playlist = media_clone.playlist.borrow_mut();

            if !playlist.is_empty() {
                let mut manager = manager_clone.lock();

                if let Some(handler) = handler(check_option!(guild_id(channel_id)), &mut manager) {
                    let first = playlist.remove(0);

                    *borrow = Some(handler.play_returning(first.source));

                    super::check_msg(channel_id.say(format!(
                        "Playing song: {}",
                        super::youtube::convert_file_name(first.name)
                    )));
                }
            }

            let two_sec = time::Duration::new(2, 0);
            thread::sleep(two_sec);
        });
    }

    fn next_song(
        song: &mut LockedAudio,
        playlist: &mut Vec<Song>,
        channel_id: ChannelId,
        manager_lock: &Arc<
            serenity::prelude::Mutex<serenity::client::bridge::voice::ClientVoiceManager>,
        >,
    ) {
        let mut manager = manager_lock.lock();

        if let Some(handler) = handler(check_option!(guild_id(channel_id)), &mut manager) {
            let first = playlist.remove(0);

            *song = handler.play_returning(first.source);

            super::check_msg(channel_id.say(format!(
                "Playing song: {}",
                super::youtube::convert_file_name(first.name)
            )));
        }
    }
}

impl Default for MediaData {
    fn default() -> MediaData {
        MediaData {
            playlist: RefCell::new(Vec::new()),
            current_song: RefCell::new(None),
        }
    }
}

unsafe impl Send for MediaData {}
unsafe impl Sync for MediaData {}

pub struct Play {
    media: Arc<MediaData>,
}

impl Play {
    pub fn new(media_data: Arc<MediaData>) -> Play {
        Play { media: media_data }
    }
}

impl serenity::framework::standard::Command for Play {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        mut ctx: &mut serenity::client::Context,
        msg: &serenity::model::channel::Message,
        mut args: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        let url = match args.single::<String>() {
            Ok(url) => url,
            Err(_) => {
                super::check_msg(msg.channel_id.say("Must provide a URL to a video or audio"));

                return Ok(());
            }
        };

        if !url.starts_with("http") {
            super::check_msg(msg.channel_id.say("Must provide a valid URL"));

            return Ok(());
        }

        let mut manager_lock = ctx.data
            .lock()
            .get::<super::VoiceManager>()
            .cloned()
            .unwrap();

        let mut source = match super::youtube::youtube_dl(&url) {
            Ok(source) => source,
            Err(why) => {
                println!("Err starting source: {:?}", why);

                super::check_msg(msg.channel_id.say("Error sourcing ffmpeg"));

                return Ok(());
            }
        };

        self.media.playlist.borrow_mut().append(&mut source);

        MediaData::start_thread(&self.media, msg.channel_id, &manager_lock);

        Ok(())
    }
}

pub struct Pause {
    media: Arc<MediaData>,
}

impl Pause {
    pub fn new(media_data: Arc<MediaData>) -> Pause {
        Pause { media: media_data }
    }
}

impl serenity::framework::standard::Command for Pause {
    #[allow(unreachable_code, unused_mut)]
    fn execute(
        &self,
        _: &mut serenity::client::Context,
        _: &serenity::model::channel::Message,
        _: serenity::framework::standard::Args,
    ) -> ::std::result::Result<(), serenity::framework::standard::CommandError> {
        if let Some(song) = self.media.current_song.borrow_mut().deref_mut() {
            //song.pause();
        }

        Ok(())
    }
}
